/*******************************************************************************
 *	BdsD.h	Produced by Bidl
 *******************************************************************************
 */
/// @file BdsD.h
/// @date 2021-05-24T09:19:37
/// @brief BOAP data class definitions for: Bds
/// @details The classes in here have been defined by a BOAP *.bidl file and
///   define classes able to be communicated across a BOAP link

#ifndef BDSD_H
#define BDSD_H 1

#include <Boap.h>
#include <BObj.h>
#include <BDate.h>
#include <BTimeStamp.h>
#include <BComplex.h>
#include <BList.h>
#include <BArray.h>

//*****************************************************************************/
//	BDS API
//			T.Barnaby,	BEAM Ltd,	2008-03-04
//*****************************************************************************/
//
namespace Bds {
	/// The System Error number list in addition to standard system error numbers.
	enum Errors 	{ ErrorNoMetaData = 64, ErrorDataQuality = 65, ErrorSlaveMode = 66, ErrorTimeStamp = 67, ErrorValidate = 80, ErrorValidateMissingBlocks = 81, ErrorValidateTimeBackwards = 82, ErrorValidateFilenameTime = 83, ErrorValidateMetaData = 84, ErrorValidateFix = 85, ErrorValidateDuplicate = 86, ErrorValidateReorder = 87, ErrorValidateBdsFudge = 88};

	/// Priority levels
	enum Priority 	{ PriorityLow, PriorityNormal, PriorityHigh};

	/// BdsServer mode
	enum Mode 	{ ModeMaster, ModeSlave};

	/// Flags when opening data files
	enum DataFlags 	{ DataFlagNone = 0x00, DataFlagClipDataToTime = 0x01, DataFlagClipDataToChannels = 0x02, DataFlagMergeSegments = 0x04, DataFlagNoMetadata = 0x08};

	/// The Selection group when making selections
	enum SelectionGroup 	{ SelectionGroupData, SelectionGroupMetaData, SelectionGroupDataWithCount};

	/// The actual format of a data sample
	enum SampleFormat 	{ SampleFormatUnknown, SampleFormatInt16, SampleFormatInt32, SampleFormatFloat32, SampleFormatFloat64, SampleFormatInt24};

	/// A flag defining the data availability state
	enum AvailType 	{ AvailNone, AvailPartial, AvailFull};

	/// Data format abilities bitset
	enum DataFormatSet 	{ DataFormatSetNone = 0x00, DataFormatSetMetadataRead = 0x01, DataFormatSetMetadataWrite = 0x02, DataFormatSetSensordataRead = 0x04, DataFormatSetSensordataWrite = 0x08};

	/// @brief This class defines an X,Y location.
	/// @details The class simply stores the x and y point values.
	class Point {
	public:
		Point(BFloat64 x = 0, BFloat64 y = 0);
	public:
		BFloat64	x;
		BFloat64	y;
	};

	/// @brief This class defines a TimePeriod.
	/// @details It has startTime and endTime fields. Note the endTime is not included in the period.
	class TimePeriod : public BObj {
	public:
		TimePeriod(BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BTimeStamp	startTime;	///< The Start time to the nearest us
		BTimeStamp	endTime;	///< The End time to the nearest us
	};

	/// @brief This class defines an integer based range.
	/// @details It is used for limit the number of items returned in selections etc.
	class ListRange : public BObj {
	public:
		ListRange(BUInt32 start = 0, BUInt32 number = 0, BInt32 reverse = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	start;	///< The start position
		BUInt32	number;	///< The number of items
		BInt32	reverse;	///< List from end
	};

	/// @brief This class defines a seismic Network organisation.
	/// @details Typical Seisomic Networks are "BN", IDC" etc.
	class Network : public BObj {
	public:
		Network(BUInt32 id = 0, BString network = BString(), BString description = BString(), BList<BString > stations = BList<BString >());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BString	network;	///< The name
		BString	description;	///< The organisations description
		BList<BString >	stations;	///< The list of arrays/stations the Network uses
	};

	/// @brief This class defines a seismic data Source.
	/// @details A Seismic data source allows different sources of data to be described and allows different Metadata sets
	/// to be used with the different data sources. It might be that there were two different digitisers in use
	/// or one data set was received real-time though a particular data processing chain while the other was via CD medium with
	/// a different processing chain.
	class Source : public BObj {
	public:
		Source(BUInt32 id = 0, BString source = BString(), BString sourceMeta = BString(), BString alias = BString(), BString description = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BString	source;	///< The sensor data's source name
		BString	sourceMeta;	///< The associated metadata's source name
		BString	alias;	///< The short alias for data files
		BString	description;	///< The description
	};

	/// @brief This class defines a Source Priority entry.
	/// @details This allows the default source for data to be selected based on a priority level.
	/// It allows a particular source to be used if no other is available and then prioritised through
	/// all the different sources available.
	class SourcePriority : public BObj {
	public:
		SourcePriority(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString source = BString(), BUInt32 priority = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BString	source;	///< The source name
		BUInt32	priority;	///< The priority order, highest first
	};

	/// @brief This class defines a full channel name.
	/// @details A channel's data is fully defined by the Network:Station:Channel:Source.
	/// This class stores all of these component names.
	class ChannelName {
	public:
		ChannelName(BString network = BString(), BString station = BString(), BString channel = BString(), BString source = BString());
	public:
		BString	network;	///< The Channels network
		BString	station;	///< The Channels station.
		BString	channel;	///< The Channels name
		BString	source;	///< The Channels source.
	};

	/// @brief This class defines an arrays channel.
	/// @details A Station could be an individual station or an array of stations.
	/// If it is an array of stations the Station object contains the list of Stations and Channels that make
	/// up the array.
	class ArrayChannel {
	public:
		ArrayChannel(BString station = BString(), BString channel = BString());
	public:
		BString	station;	///< The Stations name.
		BString	channel;	///< The Channels name
	};

	/// @brief This class defines a seismic station.
	/// @details A Station could be an individual station or an array of stations.
	/// If it is an array of stations the Station object contains the list of chaannels that make
	/// up the array.
	/// This class defines a Station
	class Station {
	public:
		Station(BUInt32 id = 0, BString name = BString(), BString alias = BString(), BString type = BString(), BString description = BString(), BList<ArrayChannel > channels = BList<ArrayChannel >());
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BString	name;	///< The name
		BString	alias;	///< Alias name to be returned to the user
		BString	type;	///< The Station type. Set to "array" or "station"
		BString	description;	///< Description
		BList<ArrayChannel >	channels;	///< List of channels if an Array.
	};

	/// @brief This class defines the physical location of a Station.
	/// @details This defines the physical location of the station using WGS84 longitude and latitude parameters.
	/// It also defines the stations elevation and if part of a seismic array, the offset with respect to the arrays centre location.
	class Location : public BObj {
	public:
		Location(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString datum = BString(), BFloat64 longitude = 0, BFloat64 latitude = 0, BFloat64 elevation = 0, BFloat64 arrayOffsetEast = 0, BFloat64 arrayOffsetNorth = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The ID
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BString	network;	///< The Network/Organisation Name
		BString	station;	///< The station this location is for
		BString	datum;	///< The locations Datum
		BFloat64	longitude;	///< The longitude in degrees using the WGS84 datum.
		BFloat64	latitude;	///< The Latitude in degrees using the WGS84 datum.
		BFloat64	elevation;	///< The ground level elevation in meters from the WGS84 ellipsoid (Sea level)
		BFloat64	arrayOffsetEast;	///< The Array offset in in an array in an easterly direction
		BFloat64	arrayOffsetNorth;	///< The Array offset in in an array in a northerly direction
	};

	//
	/// @brief This class defines a Pole/Zero Response
	/// @details It consits of an array of Complex poles and an array of Complex zeros.
	class PoleZero {
	public:
		PoleZero(BArray<BComplex > poles = BArray<BComplex >(), BArray<BComplex > zeros = BArray<BComplex >());
	public:
		BArray<BComplex >	poles;	///< Poles
		BArray<BComplex >	zeros;	///< Zeros
	};

	//
	/// @brief This class defines an entry in an Amplitude/Phase Response table
	class Fap {
	public:
		Fap(BFloat64 frequency = 0, BFloat64 amplitude = 0, BFloat64 phase = 0);
	public:
		BFloat64	frequency;	///< The frequency
		BFloat64	amplitude;	///< The Amplitude
		BFloat64	phase;	///< The Phase
	};

	//
	/// @brief This class defines an entry in a FIR coefficient table.
	class FirEntry {
	public:
		FirEntry(BFloat64 coefficient = 0, BFloat64 error = 0);
	public:
		BFloat64	coefficient;	///< Value
		BFloat64	error;	///< Error
	};

	//
	/// @brief This class defines an FIR response table.
	/// @details This has an array of the A and B coefficients.
	class Fir {
	public:
		Fir(BArray<FirEntry > b = BArray<FirEntry >(), BArray<FirEntry > a = BArray<FirEntry >());
	public:
		BArray<FirEntry >	b;	///< Numerator
		BArray<FirEntry >	a;	///< Denominator
	};

	//
	/// @brief This class defines a seismic Response characteristic.
	/// @details For each seismic channel there is a frequency response characteristic.
	/// There can be multiple stages in a channels frequency response, this response data describes one
	/// of those stages frequencies responses. The stage parameter defines which stage it is for (1, 2, 3, ...)
	/// Stage 1 is reserved to store an overall channel response.
	/// A response can be in the form of an array of poles and zeros, a FAP array, or a set of FIR coefficients.
	/// This object contains members for other responce characteristics as defined in various seismic response databases.
	class Response {
	public:
		Response(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString channel = BString(), BString source = BString(), BUInt32 stage = 0, BString name = BString(), BString type = BString(), PoleZero poleZeros = PoleZero(), BArray<Fap > faps = BArray<Fap >(), Fir fir = Fir(), BFloat64 gain = 0, BFloat64 gainFrequency = 0, BString stageType = BString(), BFloat64 decimation = 0, BString symmetry = BString(), BString description = BString(), BInt32 measured = 0, BFloat64 sampleRate = 0);
	public:
		BUInt32	id;	///< The ID
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BString	network;	///< The Network/Organisation Name
		BString	station;	///< The station
		BString	channel;	///< The channel
		BString	source;	///< The source
		BUInt32	stage;	///< The stage (0, 1, 2, 3, ...)
		BString	name;	///< The response name. (Anti-Aliasing filter, Digitiser, post filter etc)
		BString	type;	///< The type of response (PoleZero,AmplitudePhase or FIR Coefficients)
		PoleZero	poleZeros;	///< PoleZero, AmplitudePhase or FIR Coefficient data
		BArray<Fap >	faps;	///< The Frequency/Amplitude/Phase table
		Fir	fir;	///< The FIR filters coefficients
		BFloat64	gain;	///< Overall gain at gain frequency. (For information)
		BFloat64	gainFrequency;	///< Frequency that gain is valid for. (For information)
		BString	stageType;	///< The stage type: A - Analog (rad/sec), B - Analog (Hz), C - Composite, D - Digital
		BFloat64	decimation;	///< Decimation performed post filter
		BString	symmetry;	///< Symmetry for FIR coefficients (A = asymmetric, B = symmetric[odd], C = symmetric[even]) ??
		BString	description;	///< Misc description
		BInt32	measured;	///< If response was a measured response
		BFloat64	sampleRate;	///< The stage's sample rate
	};

	/// @brief This class defines a calibration setting.
	/// @details Each channel has a samplingFrequency and a calibrationFactor (scaling factor) associated with it at a particular calibrationFrequency.
	/// There may be additional calibration information such as the depth of the sensor and its positional angles.
	class Calibration : public BObj {
	public:
		Calibration(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString channel = BString(), BString source = BString(), BString name = BString(), BFloat64 samplingFrequency = 0, BFloat64 calibrationFrequency = 0, BFloat64 calibrationFactor = 0, BString calibrationUnits = BString(), BFloat64 depth = 0, BFloat64 horizontalAngle = 0, BFloat64 verticalAngle = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The ID
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BString	network;	///< The Network/Organisation Name
		BString	station;	///< The station
		BString	channel;	///< The channel
		BString	source;	///< The source
		BString	name;	///< The Calibrations name, "Main", "Measured"
		BFloat64	samplingFrequency;	///< The sample rate used in Hz
		BFloat64	calibrationFrequency;	///< The frequency that the CalibrationFactor value is valid for in Hz
		BFloat64	calibrationFactor;	///< The scaling value to apply to the data to normalise data to the units. This is a measured value at the calibration frequency.
		BString	calibrationUnits;	///< The measurment SI units such as "m"
		BFloat64	depth;	///< The depth of the sensor below ground level in meters
		BFloat64	horizontalAngle;	///< The Sensors channel placement horizontal angle in degrees clockwise from north.
		BFloat64	verticalAngle;	///< The Sensors channel placement vertical angle in degrees degrees with zero = vertically up.
	};

	/// @brief This class defines a seismic Digitiser
	/// @details This just stores information on the seismic instrument's digitiser. Its contents is generally for information only.
	class Digitiser : public BObj {
	public:
		Digitiser(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString name = BString(), BString type = BString(), BString serialNumber = BString(), BUInt32 numberChannels = 0, BFloat64 baseSamplingFrequency = 0, BFloat64 initialSamplingFrequency = 0, BFloat64 gain = 0, BInt32 shared = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The ID
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BString	name;	///< The Digitisers name
		BString	type;	///< The Digitisers type
		BString	serialNumber;	///< The digitisers's serial number.
		BUInt32	numberChannels;	///< The number of supported channels
		BFloat64	baseSamplingFrequency;	///< The base sampling frequency
		BFloat64	initialSamplingFrequency;	///< The initial pre-decimation sampling frequency
		BFloat64	gain;	///< The overall gain of the digitiser at the manufacturers calibration frequency. (For information only)
		BInt32	shared;	///< This digitiser is shared
	};

	/// @brief This class defines a seismic Sensor
	/// @details This just stores information on the seismic instrument's senosr. Its contents is generally for information only.
	class Sensor : public BObj {
	public:
		Sensor(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString name = BString(), BString type = BString(), BString serialNumber = BString(), BUInt32 numberChannels = 0, BString gainUnits = BString(), BFloat64 gain = 0, BUInt32 oldId = 0, BInt32 shared = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The ID
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BString	name;	///< The Sensors name
		BString	type;	///< The type of sensor. (Seismometer, Hydrophone etc)
		BString	serialNumber;	///< The sensor's serial number. Only used when there is a unique physical sensor.
		BUInt32	numberChannels;	///< The number of supported channels
		BString	gainUnits;	///< The gain units
		BFloat64	gain;	///< The overall gain of the sensor at the manufacturers calibration frequency. (For information only)
		BUInt32	oldId;	///< The Id from the old Autodrm database
		BInt32	shared;	///< This sensor is shared
	};

	/// @brief This class defines a Channel's instrument.
	/// @details It links a seismic data channel with a paricular sensor and digitiser.
	/// Note that it is possible to share sensor's and digitisers between channels if wanted for generic sensor/digitiser definitions.
	/// However if particular serial numbers are needed the sensor/digitiser needs to be unique.
	class ChannelInstrument : public BObj {
	public:
		ChannelInstrument(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BUInt32 channelId = 0, BString source = BString(), BUInt32 digitiserId = 0, BUInt32 sensorId = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BUInt32	channelId;	///< The channels Id
		BString	source;	///< The source
		BUInt32	digitiserId;	///< The Digitiser in use
		BUInt32	sensorId;	///< The sensor in use
	};

	/// @brief This class defines a seismic data Channel.
	/// @details This class defines a seismic data channel with network:station:channel definitions.
	/// The class also splits the channel's name field into channelType and channelAux (channel name is <channelType>_<channelAux> to]
	/// allow easy database searches etc.
	/// As well as seismic data a channel can contain other data types.
	class Channel : public BObj {
	public:
		Channel(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString channel = BString(), BString channelType = BString(), BString channelAux = BString(), BString dataType = BString(), BString description = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BString	network;	///< The Network Name
		BString	station;	///< The Stations name.
		BString	channel;	///< The channels name (often as <channelType>_<channelAux>)
		BString	channelType;	///< The channels type (component of station field)
		BString	channelAux;	///< The channels auxilary identifier (component of station field)
		BString	dataType;	///< The Type of data (seismic, seismicUnknown, data, log, unknown, empty)
		BString	description;	///< The channels description
	};

	/// @brief This class defines the set of metadata or siesmic data selected when getSelectionInfo() is use.
	/// @details This provides information on everything selected by a Selection object from the BDS metadata or seismic data sets.
	class SelectionInfo {
	public:
		SelectionInfo(BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BList<BString > networks = BList<BString >(), BList<BString > arrays = BList<BString >(), BList<BString > stations = BList<BString >(), BList<BString > arraysAndStations = BList<BString >(), BList<BString > channels = BList<BString >(), BList<BString > sources = BList<BString >(), BUInt32 numDataChannels = 0);
	public:
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BList<BString >	networks;	///< The list of Network Names
		BList<BString >	arrays;	///< The list of Array names
		BList<BString >	stations;	///< The list of Station names
		BList<BString >	arraysAndStations;	///< The list of Array and Station names
		BList<BString >	channels;	///< The list of Channels
		BList<BString >	sources;	///< The list of Data Sources
		BUInt32	numDataChannels;	///< The number of sets of data in the system matching the criteria
	};

	/// @brief This class defines a channel for selection.
	/// @details It contains the network:station:channel:source names.
	class SelectionChannel {
	public:
		SelectionChannel(BString network = BString(), BString station = BString(), BString channel = BString(), BString source = BString());
	public:
		BString	network;
		BString	station;
		BString	channel;
		BString	source;
	};

	/// @brief This class defines a generic metadata or seismic data selection.
	/// @details This defines a set of selection criteria when selecting items from the BDS metadata or seismic data sets.
	/// The fields, when set, limit the ietms returned by the settings provided. In effect it has an "AND" type of function
	/// with the parameters provided.
	/// There are some specific parameters like: channelId, digitiserId, sensorId etc which are used with specific
	/// data selection functions.
	class Selection {
	public:
		Selection(BUInt32 id = 0, ListRange range = ListRange(), BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BList<SelectionChannel > channels = BList<SelectionChannel >(), BUInt32 channelId = 0, BUInt32 digitiserId = 0, BUInt32 sensorId = 0, BUInt32 sensorOldId = 0, BInt32 completeSegments = 0, BString calibrationName = BString());
	public:
		BUInt32	id;	///< The ID of the record to return
		ListRange	range;	///< The range of data to return
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BList<SelectionChannel >	channels;	///< The data channels to select
		BUInt32	channelId;	///< The Channel id
		BUInt32	digitiserId;	///< The Digitiser id
		BUInt32	sensorId;	///< The Sensor id
		BUInt32	sensorOldId;	///< The Sensor old id
		BInt32	completeSegments;	///< Do not clip the segment times to match the required time period
		BString	calibrationName;	///< Calibration name to use
	};

	/// @brief This class provides information on a channel
	/// @details This returns the metadata available for a channel over a particular time period.
	/// There are likely to be multiple ChannelInfo objects over larger time periods, one for each change in metadata.
	/// The ChannelInfos object contains an array of these ChannelInfo objects.
	class ChannelInfo {
	public:
		ChannelInfo(BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), Station station = Station(), Location location = Location(), Channel channel = Channel(), BString source = BString(), BString dataType = BString(), Digitiser digitiser = Digitiser(), Sensor sensor = Sensor(), Calibration calibration = Calibration(), BList<Response > responses = BList<Response >());
	public:
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		Station	station;	///< The Station info
		Location	location;	///< The Station location
		Channel	channel;	///< The Channel data
		BString	source;	///< The data source
		BString	dataType;	///< The DataType (seismic, seismicUnknown, data, log, unknown, empty)
		Digitiser	digitiser;	///< The Digitiser in use
		Sensor	sensor;	///< The Sensor in use
		Calibration	calibration;	///< The Calibration info
		BList<Response >	responses;	///< The list of frequency responses
	};

	/// @brief This class provides metadata information on a set of channels
	/// @details This provides all of the metadata information for a set of channels over a given time period.
	/// For each channel there can be one or more ChannelInfo objects each defining a set of consistant metadata
	/// over a particular time period. The BdsServer will create a ChannelInfo object on each cahnge in metadata
	/// when the user asks for a set of metadata over some time period.
	class ChannelInfos {
	public:
		ChannelInfos(BArray<BArray<ChannelInfo > > channels = BArray<BArray<ChannelInfo > >());
	public:
		BArray<BArray<ChannelInfo > >	channels;
	};

	/// @brief This class defines information on a sensor data file
	/// @details The raw sensor data for a seismic channel is stored in files in the BDS system.
	/// This class defines the database entry that describes this file, its storage location and status.
	/// A single file can store one or more channels of seismic data in different formats.
	class DataFileInfo : public BObj {
	public:
		DataFileInfo(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString location = BString(), BString format = BString(), BString url = BString(), BString comment = BString(), BUInt32 importUserId = 0, BTimeStamp importTime = BTimeStamp(), BString state = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BString	location;	///< The storage location
		BString	format;	///< The data format
		BString	url;	///< The URL for file access
		BString	comment;	///< A comment on the file
		BUInt32	importUserId;	///< The user ID of the importing user
		BTimeStamp	importTime;	///< The Time the data was imported
		BString	state;	///< Status info on the import (importing, realtime, failed, ok etc)
	};

	/// @brief This class defines information on a single channels set of data stored in a file.
	/// @details This provides information on actual seismic data for a channel that is stored in the BDS system.
	/// The data will be stored in a particular file perhaps with other data channels.
	/// When known information on the channels numBlocks, numSamples and sampleRate will be provided. Generally this information
	/// will only be know if a data files has been imported rather than a live real-time data stream.
	/// Generally the seismic data file itself should be interrogated to find the definitive information.
	/// The info field provides extra details on the data contents which might come from one of the spefic data import formats.
	class DataChannel : public BObj {
	public:
		DataChannel(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString channel = BString(), BString source = BString(), BUInt32 numBlocks = 0, BUInt64 numSamples = 0, BFloat64 sampleRate = 0, BUInt32 sampleFormat = 0, BUInt32 dataFileId = 0, BUInt32 dataFileChannel = 0, BString importFormat = BString(), BString importFilename = BString(), BTimeStamp importStartTime = BTimeStamp(), BDict<BString > info = BDict<BString >());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BString	network;	///< The Network Name
		BString	station;	///< The Station name
		BString	channel;	///< The Channels name
		BString	source;	///< The Data Source
		BUInt32	numBlocks;	///< The total number of blocks per channel if known, 0 otherwise
		BUInt64	numSamples;	///< The total number of samples per channel if known, 0 otherwise
		BFloat64	sampleRate;	///< The data's sample rate
		BUInt32	sampleFormat;	///< The data sample format
		BUInt32	dataFileId;	///< The Data File Id. This links to the particular DataFileInfo where the data is stored.
		BUInt32	dataFileChannel;	///< The Data File Channel number. The channel number within the data file. (1, 2, 3 ...)
		BString	importFormat;	///< The original data format
		BString	importFilename;	///< The original data file name
		BTimeStamp	importStartTime;	///< The original import files start time
		BDict<BString >	info;	///< Extra info on the channel
	};

	/// @brief This class defines information on a set of data.
	/// @details This describes a set of seismic data. It is returns basic informatuion when performing a selection of data
	/// or detailed information when enquiring information from an actual seismic data file.
	/// All of the detailed information comes from the data files themselves. This includes the info, infoExtra and warnings information.
	/// The channels array contains an array of data segments per channel. For basic information this could be a single segment over
	/// a time period. However, when equiring detailed information from a file it will contain an entry per contiguous data segemnt
	/// in the file.
	class DataInfo {
	public:
		DataInfo(BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString array = BString(), BString description = BString(), BUInt32 synchronous = 0, BArray<BArray<DataChannel > > channels = BArray<BArray<DataChannel > >(), BDict<BString > info = BDict<BString >(), BDict<BString > infoExtra = BDict<BString >(), BList<BString > warnings = BList<BString >());
	public:
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BString	array;	///< The Seismic Array that all of the channels are from, if just one
		BString	description;	///< The Comment
		BUInt32	synchronous;	///< The channels are synchronously sampled
		BArray<BArray<DataChannel > >	channels;	///< The Data channels. Each channel can have multiple segments of data
		BDict<BString >	info;	///< Info on the set of channels
		BDict<BString >	infoExtra;	///< Extra Info on the set of channels. Used for extended error/logging information
		BList<BString >	warnings;	///< Warnings on the data set
	};

	/// @brief This class provides availability information on a particular period of data.
	class DataAvail {
	public:
		DataAvail(BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), AvailType availType = AvailType());
	public:
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		AvailType	availType;	///< The availability type. Can be: AvailNone, AvailPartial or AvailFull.
	};

	/// @brief This class defines avaiilablity information on a set of data.
	/// @details Availability information is provided for a particular channel network:station:channel:source.
	/// For a particular time period this will be broken down into time period segments.
	/// The time granularity of the segments is dependent of the type of data availability search.
	/// Normally the granularity is that defined by the DataChannel information in the database. The actual seismic data files
	/// may not contains some data over the given periods due to missing blocks etc.
	/// An in-depth adta availability search could provide more fuller data availability information but with much more data processing.
	/// This has yet to be implemented.
	class DataAvailChan {
	public:
		DataAvailChan(BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString channel = BString(), BString source = BString(), BArray<DataAvail > segments = BArray<DataAvail >());
	public:
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BString	network;	///< The Network Name
		BString	station;	///< The Station name
		BString	channel;	///< The Channels name
		BString	source;	///< The Data Source
		BArray<DataAvail >	segments;	///< Segment info
	};

	/// @brief This defines a handle to a sensor data stream/file when opened for read or write.
	class DataHandle {
	public:
		DataHandle(BUInt32 handle = 0, BUInt32 dataFileId = 0);
	public:
		BUInt32	handle;	///< Opaque file handle
		BUInt32	dataFileId;	///< The data file ID if opened for write
	};

	/// @brief This class provides the acual seismic data values contained within a single data block.
	/// @details All seismic data is segmented into blocks that have a timestamp and perhaps other metadata.
	/// The DataBlock contains an array of actual data samples in a 64bit floating pointer format for one or more channels.
	/// If the samples are for multiple channels it is assumed that these are synchronously sampled and are normally from
	/// a data file that is in the sample multiplexed format (ie. For each point in time there is a set of samples one for each cahnnel).
	/// The startTime comes from the original blocks start time information.
	/// The endTime may come from the original blocks end time information, if available in the original data format that the data was
	/// imported from or is generated from the startTime and the calculated sample rate of the data.
	/// For some data types where the sampling rate is a bit varible, the endTime fields may be lined up with the next blocks startTime
	/// field to ensure contiguous data segments.
	/// The info field contains extra, free string format, metadata on the block if available.
	/// This could be quality information from the TapeDigitiser system for example.
	class DataBlock {
	public:
		DataBlock(BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BUInt32 channelNumber = 0, BUInt32 segmentNumber = 0, BArray<BArray<BFloat64 > > channelData = BArray<BArray<BFloat64 > >(), BDict<BString > info = BDict<BString >());
	public:
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BUInt32	channelNumber;	///< The first channel number. (1, 2, 3 ...)
		BUInt32	segmentNumber;	///< The segment number. (1, 2, 3, ...)
		BArray<BArray<BFloat64 > >	channelData;	///< The raw channel data in a 2 dimensional array, ordered as per channel information in dataInfo
		BDict<BString >	info;	///< Extra information on data or ASCII data
	};

	/// @brief This holds information on a user
	/// @details All information on a BDS users is stored along with the security groups they belong to.
	class User : public BObj {
	public:
		User(BUInt32 id = 0, BString user = BString(), BString password = BString(), BString name = BString(), BString email = BString(), BString telephone = BString(), BString address = BString(), BInt32 enabled = 0, BList<BString > groups = BList<BString >());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The unique user ID
		BString	user;	///< The User ID
		BString	password;	///< The Users password
		BString	name;	///< The Users full name
		BString	email;	///< The users email Address
		BString	telephone;	///< The Users telephone number
		BString	address;	///< The Users postal address
		BInt32	enabled;	///< Whether the users account is enabled
		BList<BString >	groups;	///< The security groups the user belongs to
	};

	/// @brief This holds information on a user security group.
	/// @details The BDS has the concept of a security group that users can belong to.
	/// This class defines that security group as stored in the database.
	class Group : public BObj {
	public:
		Group(BUInt32 id = 0, BString group = BString(), BString description = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The unique id
		BString	group;	///< The Group name
		BString	description;	///< The Groups description
	};

	/// @brief This holds information on data access groups.
	/// @details A particular network:station may contain sensitive data. This database linked object
	/// links a period of data from a particular network:station to a security group.
	class AccessGroup : public BObj {
	public:
		AccessGroup(BUInt32 id = 0, BString group = BString(), BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The unique id
		BString	group;	///< The Group name
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time
		BString	network;	///< The Network Name
		BString	station;	///< The Station name
	};

	/// @brief This holds information on a medatdata or sensor data change
	/// @details Whenever a change is made to the BDS metadata or data a Change object is added to the BDS Changes database.
	/// This describes which database table and object that was added or modified.
	/// Change's are normally grouped together by a ChangeGroup.
	class Change : public BObj {
	public:
		Change(BUInt32 id = 0, BUInt32 changeGroupId = 0, BTimeStamp time = BTimeStamp(), BString type = BString(), BString table = BString(), BUInt32 rowId = 0);
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The unique id
		BUInt32	changeGroupId;	///< The Change group ID
		BTimeStamp	time;	///< The Time the change was made
		BString	type;	///< The change type
		BString	table;	///< The database table affected
		BUInt32	rowId;	///< The database row affected
	};

	/// @brief This holds information on a set of Changes
	/// @details A set of changes to the BDS database are grouped into a ChangeGroup.
	/// This could be a set of changes whilst a user is logged in or by a program making
	/// a set of changes.
	class ChangeGroup : public BObj {
	public:
		ChangeGroup(BUInt32 id = 0, BTimeStamp time = BTimeStamp(), BString type = BString(), BString user = BString(), BString title = BString(), BString description = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The unique id
		BTimeStamp	time;	///< The Time the change was made
		BString	type;	///< The type of change
		BString	user;	///< The user who made the change
		BString	title;	///< The Changes title
		BString	description;	///< The Description of the change
	};

	/// @brief This holds information on a Note for general information
	/// @details Normally a Note can be added for a particular set of data over a particular time period.
	/// These notes are sometimes added automaically during a data import process or by a user.
	/// A data user can then ask for any notes for a particular set of data.
	class Note : public BObj {
	public:
		Note(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString channel = BString(), BString source = BString(), BString type = BString(), BString user = BString(), BTimeStamp timeAdded = BTimeStamp(), BInt32 errorNumber = 0, BString title = BString(), BString description = BString(), BString docFormat = BString(), BString docUrl = BString(), BUInt32 dataFileId = 0, BString importFilename = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The unique id
		BTimeStamp	startTime;	///< The Start Time note is for
		BTimeStamp	endTime;	///< The End Time note is for
		BString	network;	///< The Network Name
		BString	station;	///< The Station/Array name
		BString	channel;	///< The Channels name
		BString	source;	///< The Data Source
		BString	type;	///< The Type (note, warning, error ...)
		BString	user;	///< The user
		BTimeStamp	timeAdded;	///< The Time Entered
		BInt32	errorNumber;	///< Error number if error
		BString	title;	///< The title
		BString	description;	///< The Description
		BString	docFormat;	///< Document format if any
		BString	docUrl;	///< Document Url if any
		BUInt32	dataFileId;	///< The data file id associated with this note
		BString	importFilename;	///< The import filename
	};

	/// @brief This holds information on a Log entry.
	/// @details Log entries are added automatically and manually to the system.
	/// A system administartor can view these logs.
	class Log : public BObj {
	public:
		Log(BUInt32 id = 0, BTimeStamp time = BTimeStamp(), BString type = BString(), BUInt32 priority = 0, BString subSystem = BString(), BString title = BString(), BString description = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< The unique id
		BTimeStamp	time;	///< The Time
		BString	type;	///< The Type
		BUInt32	priority;	///< The priority 0 to 5
		BString	subSystem;	///< The SubSystem
		BString	title;	///< The Changes title
		BString	description;	///< The Description of the change
	};

	/// @brief This defines the selection cirteria when requesting a set of log entries.
	class LogSelect {
	public:
		LogSelect(BTimeStamp startTime = BTimeStamp(), BString type = BString(), BUInt32 priority = 0, BString subSystem = BString());
	public:
		BTimeStamp	startTime;	///< The start time
		BString	type;	///< The Type
		BUInt32	priority;	///< The priority 0 to 5
		BString	subSystem;	///< The SubSystem
	};

	/// @brief This defines the set of clean options used in the clean() function.
	class CleanOptions {
	public:
		CleanOptions(BInt32 logs = 0, BInt32 changes = 0, BInt32 deletedFiles = 0);
	public:
		BInt32	logs;	///< Clean the Logs
		BInt32	changes;	///< Clean the changes
		BInt32	deletedFiles;	///< Clean deleted data files
	};

	/// @brief This holds information on a seismic data format.
	/// @details It is used by the BDS data convertors to define which data formats they support..
	class DataFormat {
	public:
		DataFormat(BList<BString > names = BList<BString >(), BString description = BString(), BInt32 dataRead = 0, BInt32 dataWrite = 0, BInt32 metaDataRead = 0, BInt32 metaDataWrite = 0, BString extension = BString());
	public:
		BList<BString >	names;	///< The format names
		BString	description;	///< The description
		BInt32	dataRead;	///< Ability to read data
		BInt32	dataWrite;	///< Ability to write data
		BInt32	metaDataRead;	///< MetaData read supported
		BInt32	metaDataWrite;	///< MetaData write supported
		BString	extension;	///< Default filename extension
	};

	class SpecialChannel : public BObj {
	public:
		SpecialChannel(BUInt32 id = 0, BTimeStamp startTime = BTimeStamp(), BTimeStamp endTime = BTimeStamp(), BString network = BString(), BString station = BString(), BString channel = BString(), BString type = BString(), BString description = BString());
		BString getType();
		BError setMembers(BDictString& members);
		BError setMember(BString name, BString value);
		BError getMembers(BDictString& members);
		BError getMember(BString name, BString& value);
	public:
		BUInt32	id;	///< Unique ID when stored in a database or for other uses.
		BTimeStamp	startTime;	///< The Start Time
		BTimeStamp	endTime;	///< The End Time the channel was available
		BString	network;	///< The Network Name, wildcards allowed
		BString	station;	///< The Stations name, wildcards allowed.
		BString	channel;	///< The channels name, wildcards allowed (often as <channelType>_<channelAux>)
		BString	type;	///< The Type of channel (ignore)
		BString	description;	///< The channels description
	};

	/// @class DataAccess
	/// @brief This is the Data Access API interface.
	/// @details This object provides the set of API functions that make RPC network calls to a BdsServer.
	/// It provides the basic and restricted user orientated read only, data access API.
	/// @class DataAddAccess
	/// @brief This is the DataAdd Access API interface.
	/// @details This object provides the set of API functions that make RPC network calls to a BdsServer.
	/// It provides the basic and restricted user orientated read only, data access API along with the ability
	/// to import data. It will normally be used by data import client programs.
	/// @class AdminAccess
	/// @brief This is the AdminAccess Access API interface.
	/// @details This object provides the set of API functions that make RPC network calls to a BdsServer.
	/// It provides the full unresiticed data access API allowring all data read and write operations to be performed.
}

#endif
